# mypy: disable-error-code="annotation-unchecked"
import copy
import dataclasses
import functools
import itertools
import threading
import unittest
import uuid
from datetime import datetime
from enum import Enum
from typing import Any, NamedTuple, Optional
from unittest.mock import MagicMock, patch

import attr
import dataclasses_json
import pytest
from pydantic import BaseModel

import langsmith.utils as ls_utils
from langsmith import Client, traceable
from langsmith.run_helpers import get_current_run_tree, tracing_context


class LangSmithProjectNameTest(unittest.TestCase):
    class GetTracerProjectTestCase:
        def __init__(
            self, test_name, envvars, expected_project_name, return_default_value=None
        ):
            self.test_name = test_name
            self.envvars = envvars
            self.expected_project_name = expected_project_name
            self.return_default_value = return_default_value

    def test_correct_get_tracer_project(self):
        ls_utils.get_env_var.cache_clear()
        cases = [
            self.GetTracerProjectTestCase(
                test_name="default to 'default' when no project provided",
                envvars={},
                expected_project_name="default",
            ),
            self.GetTracerProjectTestCase(
                test_name="default to 'default' when "
                + "return_default_value=True and no project provided",
                envvars={},
                expected_project_name="default",
            ),
            self.GetTracerProjectTestCase(
                test_name="do not default if return_default_value=False "
                + "when no project provided",
                envvars={},
                expected_project_name=None,
                return_default_value=False,
            ),
            self.GetTracerProjectTestCase(
                test_name="use session_name for legacy tracers",
                envvars={"LANGCHAIN_SESSION": "old_timey_session"},
                expected_project_name="old_timey_session",
            ),
            self.GetTracerProjectTestCase(
                test_name="use LANGCHAIN_PROJECT over SESSION_NAME",
                envvars={
                    "LANGCHAIN_SESSION": "old_timey_session",
                    "LANGCHAIN_PROJECT": "modern_session",
                },
                expected_project_name="modern_session",
            ),
            self.GetTracerProjectTestCase(
                test_name="hosted projects get precedence over all other defaults",
                envvars={
                    "HOSTED_LANGSERVE_PROJECT_NAME": "hosted_project",
                    "LANGCHAIN_SESSION": "old_timey_session",
                    "LANGCHAIN_PROJECT": "modern_session",
                },
                expected_project_name="hosted_project",
            ),
        ]

        for case in cases:
            ls_utils.get_env_var.cache_clear()
            ls_utils.get_tracer_project.cache_clear()
            with self.subTest(msg=case.test_name):
                with pytest.MonkeyPatch.context() as mp:
                    for k, v in case.envvars.items():
                        mp.setenv(k, v)

                    project = (
                        ls_utils.get_tracer_project()
                        if case.return_default_value is None
                        else ls_utils.get_tracer_project(case.return_default_value)
                    )
                    self.assertEqual(project, case.expected_project_name)


def test_tracing_enabled():
    ls_utils.get_env_var.cache_clear()
    with patch.dict(
        "os.environ", {"LANGCHAIN_TRACING_V2": "false", "LANGSMITH_TRACING": "false"}
    ):
        assert not ls_utils.tracing_is_enabled()
        with tracing_context(enabled=True):
            assert ls_utils.tracing_is_enabled()
            with tracing_context(enabled=False):
                assert not ls_utils.tracing_is_enabled()
        with tracing_context(enabled=False):
            assert not ls_utils.tracing_is_enabled()
        assert not ls_utils.tracing_is_enabled()

    @traceable
    def grandchild_function():
        assert ls_utils.tracing_is_enabled()
        rt = get_current_run_tree()
        assert rt
        assert rt.parent_run_id is None
        assert "." not in rt.dotted_order
        assert rt.parent_dotted_order is None
        return 1

    @traceable
    def child_function():
        assert ls_utils.tracing_is_enabled()
        with tracing_context(parent=False):
            return grandchild_function()

    @traceable
    def untraced_child_function():
        assert not ls_utils.tracing_is_enabled()
        return 1

    @traceable
    def parent_function():
        with patch.dict(
            "os.environ",
            {"LANGCHAIN_TRACING_V2": "false", "LANGSMITH_TRACING": "false"},
        ):
            assert ls_utils.tracing_is_enabled()
            child_function()
        with tracing_context(enabled=False):
            assert not ls_utils.tracing_is_enabled()
            return untraced_child_function()

    ls_utils.get_env_var.cache_clear()
    with patch.dict(
        "os.environ", {"LANGCHAIN_TRACING_V2": "true", "LANGSMITH_TRACING": "true"}
    ):
        mock_client = MagicMock(spec=Client)
        parent_function(langsmith_extra={"client": mock_client})


def test_tracing_disabled():
    ls_utils.get_env_var.cache_clear()
    with patch.dict(
        "os.environ", {"LANGCHAIN_TRACING_V2": "true", "LANGSMITH_TRACING": "true"}
    ):
        assert ls_utils.tracing_is_enabled()
        with tracing_context(enabled=False):
            assert not ls_utils.tracing_is_enabled()
        with tracing_context(enabled=True):
            assert ls_utils.tracing_is_enabled()
            with tracing_context(enabled=False):
                assert not ls_utils.tracing_is_enabled()
        assert ls_utils.tracing_is_enabled()


def test_deepish_copy():
    class MyClass:
        def __init__(self, x: int) -> None:
            self.x = x
            self.y = "y"
            self.a_list = [1, 2, 3]
            self.a_tuple = (1, 2, 3)
            self.a_set = {1, 2, 3}
            self.a_dict = {"foo": "bar"}
            self.my_bytes = b"foo"

    class ClassWithTee:
        def __init__(self) -> None:
            tee_a, tee_b = itertools.tee(range(10))
            self.tee_a = tee_a
            self.tee_b = tee_b

    class MyClassWithSlots:
        __slots__ = ["x", "y"]

        def __init__(self, x: int) -> None:
            self.x = x
            self.y = "y"

    class MyPydantic(BaseModel):
        foo: str
        bar: int
        baz: dict

    @dataclasses.dataclass
    class MyDataclass:
        foo: str
        bar: int

        def something(self) -> None:
            pass

    class MyEnum(str, Enum):
        FOO = "foo"
        BAR = "bar"

    class ClassWithFakeJson:
        def json(self):
            raise ValueError("This should not be called")

        def to_json(self) -> dict:
            return {"foo": "bar"}

    @dataclasses_json.dataclass_json
    @dataclasses.dataclass
    class Person:
        name: str

    @attr.dataclass
    class AttrDict:
        foo: str = attr.ib()
        bar: int

    uid = uuid.uuid4()
    current_time = datetime.now()

    class NestedClass:
        __slots__ = ["person", "lock"]

        def __init__(self) -> None:
            self.person = Person(name="foo")
            self.lock = [threading.Lock()]

        def __deepcopy__(self, memo: Optional[dict] = None) -> Any:
            cls = type(self)
            m = cls.__new__(cls)
            setattr(m, "__dict__", copy.deepcopy(self.__dict__, memo=memo))

    class CyclicClass:
        def __init__(self) -> None:
            self.cyclic = self

        def __repr__(self) -> str:
            return "SoCyclic"

    class CyclicClass2:
        def __init__(self) -> None:
            self.cyclic: Any = None
            self.other: Any = None

        def __repr__(self) -> str:
            return "SoCyclic2"

    cycle_2 = CyclicClass2()
    cycle_2.cyclic = CyclicClass2()
    cycle_2.cyclic.other = cycle_2

    class MyNamedTuple(NamedTuple):
        foo: str
        bar: int

    my_dict = {
        "uid": uid,
        "time": current_time,
        "adict": {"foo": "bar"},
        "my_class": MyClass(1),
        "class_with_tee": ClassWithTee(),
        "my_slotted_class": MyClassWithSlots(1),
        "my_dataclass": MyDataclass("foo", 1),
        "my_enum": MyEnum.FOO,
        "my_pydantic": MyPydantic(foo="foo", bar=1, baz={"foo": "bar"}),
        "person": Person(name="foo"),
        "a_bool": True,
        "a_none": None,
        "a_str": "foo",
        "an_int": 1,
        "a_float": 1.1,
        "nested_class": NestedClass(),
        "attr_dict": AttrDict(foo="foo", bar=1),
        "named_tuple": MyNamedTuple(foo="foo", bar=1),
        "cyclic": CyclicClass(),
        "cyclic2": cycle_2,
        "fake_json": ClassWithFakeJson(),
    }
    assert ls_utils.deepish_copy(my_dict) == my_dict


def test_is_version_greater_or_equal():
    # Test versions equal to 0.5.23
    assert ls_utils.is_version_greater_or_equal("0.5.23", "0.5.23")

    # Test versions greater than 0.5.23
    assert ls_utils.is_version_greater_or_equal("0.5.24", "0.5.23")
    assert ls_utils.is_version_greater_or_equal("0.6.0", "0.5.23")
    assert ls_utils.is_version_greater_or_equal("1.0.0", "0.5.23")

    # Test versions less than 0.5.23
    assert not ls_utils.is_version_greater_or_equal("0.5.22", "0.5.23")
    assert not ls_utils.is_version_greater_or_equal("0.5.0", "0.5.23")
    assert not ls_utils.is_version_greater_or_equal("0.4.99", "0.5.23")


def test_parse_prompt_identifier():
    # Valid cases
    assert ls_utils.parse_prompt_identifier("name") == ("-", "name", "latest")
    assert ls_utils.parse_prompt_identifier("owner/name") == ("owner", "name", "latest")
    assert ls_utils.parse_prompt_identifier("owner/name:commit") == (
        "owner",
        "name",
        "commit",
    )
    assert ls_utils.parse_prompt_identifier("name:commit") == ("-", "name", "commit")

    # Invalid cases
    invalid_identifiers = [
        "",
        "/",
        ":",
        "owner/",
        "/name",
        "owner//name",
        "owner/name/",
        "owner/name/extra",
        ":commit",
    ]

    for invalid_id in invalid_identifiers:
        try:
            ls_utils.parse_prompt_identifier(invalid_id)
            assert False, f"Expected ValueError for identifier: {invalid_id}"
        except ValueError:
            pass  # This is the expected behavior


def test_get_api_key() -> None:
    ls_utils.get_env_var.cache_clear()
    assert ls_utils.get_api_key("provided_api_key") == "provided_api_key"
    assert ls_utils.get_api_key("'provided_api_key'") == "provided_api_key"
    assert ls_utils.get_api_key('"_provided_api_key"') == "_provided_api_key"

    with patch.dict("os.environ", {"LANGCHAIN_API_KEY": "env_api_key"}, clear=True):
        api_key_ = ls_utils.get_api_key(None)
        assert api_key_ == "env_api_key"

    ls_utils.get_env_var.cache_clear()

    with patch.dict("os.environ", {}, clear=True):
        assert ls_utils.get_api_key(None) is None
    ls_utils.get_env_var.cache_clear()
    assert ls_utils.get_api_key("") is None
    assert ls_utils.get_api_key(" ") is None


def test_get_api_url() -> None:
    ls_utils.get_env_var.cache_clear()
    assert ls_utils.get_api_url("http://provided.url") == "http://provided.url"

    with patch.dict("os.environ", {"LANGCHAIN_ENDPOINT": "http://env.url"}):
        assert ls_utils.get_api_url(None) == "http://env.url"

    ls_utils.get_env_var.cache_clear()
    with patch.dict("os.environ", {}, clear=True):
        assert ls_utils.get_api_url(None) == "https://api.smith.langchain.com"
    ls_utils.get_env_var.cache_clear()
    with patch.dict("os.environ", {}, clear=True):
        assert ls_utils.get_api_url(None) == "https://api.smith.langchain.com"
    ls_utils.get_env_var.cache_clear()
    with patch.dict("os.environ", {"LANGCHAIN_ENDPOINT": "http://env.url"}):
        assert ls_utils.get_api_url(None) == "http://env.url"
    ls_utils.get_env_var.cache_clear()
    with pytest.raises(ls_utils.LangSmithUserError):
        ls_utils.get_api_url(" ")


def test_get_func_name():
    class Foo:
        def __call__(self, foo: int):
            return "bar"

    assert ls_utils._get_function_name(Foo()) == "Foo"
    assert ls_utils._get_function_name(functools.partial(Foo(), foo=3)) == "Foo"

    class AFoo:
        async def __call__(self, foo: int):
            return "bar"

    assert ls_utils._get_function_name(AFoo()) == "AFoo"
    assert ls_utils._get_function_name(functools.partial(AFoo(), foo=3)) == "AFoo"

    def foo(bar: int) -> None:
        return bar

    assert ls_utils._get_function_name(foo) == "foo"
    assert ls_utils._get_function_name(functools.partial(foo, bar=3)) == "foo"

    async def afoo(bar: int) -> None:
        return bar

    assert ls_utils._get_function_name(afoo) == "afoo"
    assert ls_utils._get_function_name(functools.partial(afoo, bar=3)) == "afoo"

    lambda_func = lambda x: x + 1  # noqa
    assert ls_utils._get_function_name(lambda_func) == "<lambda>"

    class BarClass:
        pass

    assert ls_utils._get_function_name(BarClass) == "BarClass"

    assert ls_utils._get_function_name(print) == "print"

    assert ls_utils._get_function_name("not_a_function") == "not_a_function"


def test_get_host_url():
    # If web_url is explicitly provided, it takes precedence over api_url.
    assert (
        ls_utils.get_host_url(
            "https://my-custom-web.com", "https://api.smith.langchain.com"
        )
        == "https://my-custom-web.com"
    )

    # When web_url is None and api_url is localhost.
    assert ls_utils.get_host_url(None, "http://localhost:5000") == "http://localhost"
    # A port variation on localhost.
    assert ls_utils.get_host_url(None, "http://127.0.0.1:8080") == "http://localhost", (
        "Should recognize 127.x.x.x as localhost."
    )

    # If api_url path ends with /api, trimmed back to netloc.
    assert (
        ls_utils.get_host_url(None, "https://my-awesome-domain.com/api")
        == "https://my-awesome-domain.com"
    )

    # If api_url path ends with /api/v1, trimmed back to netloc.
    assert (
        ls_utils.get_host_url(None, "https://my-other-domain.com/api/v1")
        == "https://my-other-domain.com"
    )

    # If netloc begins with dev.
    assert (
        ls_utils.get_host_url(None, "https://dev.smith.langchain.com/api/v1")
        == "https://dev.smith.langchain.com"
    )

    # If netloc begins with eu.
    assert (
        ls_utils.get_host_url(None, "https://eu.smith.langchain.com/api")
        == "https://eu.smith.langchain.com"
    )

    # If netloc begins with beta.
    assert (
        ls_utils.get_host_url(None, "https://beta.smith.langchain.com")
        == "https://beta.smith.langchain.com"
    )

    # If netloc begins with api.
    assert (
        ls_utils.get_host_url(None, "https://api.smith.langchain.com")
        == "https://smith.langchain.com"
    )

    # Otherwise, returns https://smith.langchain.com for unknown host.
    assert (
        ls_utils.get_host_url(None, "https://unknownhost.com")
        == "https://smith.langchain.com"
    )


class MockRequest:
    """Mock request object for testing filter_request_headers"""

    def __init__(self, url: str, headers: Optional[dict] = None):
        self.url = url
        self.headers = headers or {
            "Authorization": "Bearer token",
            "Content-Type": "application/json",
        }


def test_filter_request_headers():
    # Test with no filtering - both ignore_hosts and allow_hosts are None
    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(request)

    assert result is request


def test_filter_request_headers_allow_hosts():
    # Test allow_hosts functionality

    # Test matching hostname (exact match)
    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["api.example.com"])
    assert result is request

    # Test matching full URL
    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["https://api.example.com"]
    )
    assert result is request

    # Test subdomain matching
    request = MockRequest("https://sub.example.com/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["example.com"])
    assert result is request

    # Test non-matching host - should return None
    request = MockRequest("https://api.different.com/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["api.example.com"])
    assert result is None

    # Test multiple allow hosts - matching first
    request = MockRequest("https://api.first.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["api.first.com", "api.second.com"]
    )
    assert result is request

    # Test multiple allow hosts - matching second
    request = MockRequest("https://api.second.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["api.first.com", "api.second.com"]
    )
    assert result is request

    # Test multiple allow hosts - not matching any
    request = MockRequest("https://api.third.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["api.first.com", "api.second.com"]
    )
    assert result is None


def test_filter_request_headers_url_parsing_error():
    # Test URL parsing error handling
    class BadRequest:
        def __init__(self, bad_url: str):
            self.url = bad_url
            self.headers = {"Authorization": "Bearer token"}

    # Mock urllib_parse.urlparse to raise an exception
    with patch(
        "langsmith.utils.urllib_parse.urlparse", side_effect=Exception("Parse error")
    ):
        request = BadRequest("not-a-valid-url")
        result = ls_utils.filter_request_headers(request, allow_hosts=["example.com"])
        assert result is None


def test_filter_request_headers_hostname_none():
    # Test when parsed_url.hostname is None
    request = MockRequest("file:///local/path")  # This URL has no hostname
    result = ls_utils.filter_request_headers(request, allow_hosts=["example.com"])
    assert result is None


def test_filter_request_headers_combined_ignore_and_allow():
    # Test when both ignore_hosts and allow_hosts are provided
    # ignore_hosts takes precedence (legacy behavior)

    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(
        request,
        ignore_hosts=["https://api.example.com"],
        allow_hosts=["api.example.com"],
    )
    assert result is None  # Should be ignored despite being in allow_hosts


def test_filter_request_headers_edge_cases():
    # Test various edge cases

    # Empty ignore_hosts list
    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(request, ignore_hosts=[])
    assert result is request

    # Empty allow_hosts list - empty list evaluates to False, so no filtering occurs
    request = MockRequest("https://api.example.com/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=[])
    assert result is request  # Empty list means no allow_hosts filtering

    # HTTP vs HTTPS in allow_hosts
    request = MockRequest("http://api.example.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["https://api.example.com"]
    )
    assert result is None  # Should not match different protocol

    # Port numbers in URLs
    request = MockRequest("https://api.example.com:8080/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["api.example.com"])
    assert result is request

    # Subdomain with allow_hosts containing full URL
    request = MockRequest("https://sub.example.com/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["https://example.com"]
    )
    assert result is None  # Full URL should not match subdomain


def test_filter_request_headers_localhost():
    # Test localhost with different ports

    # Test localhost with port 8080
    request = MockRequest("http://localhost:8080/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["localhost"])
    assert result is request

    # Test 127.0.0.1 with port 3000
    request = MockRequest("http://127.0.0.1:3000/api")
    result = ls_utils.filter_request_headers(request, allow_hosts=["127.0.0.1"])
    assert result is request

    # Test localhost without port matching localhost with port
    request = MockRequest("http://localhost/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["localhost"])
    assert result is request

    # Test different localhost formats
    request = MockRequest("https://127.0.0.1:5000/test")
    result = ls_utils.filter_request_headers(request, allow_hosts=["localhost"])
    assert result is None  # 127.0.0.1 doesn't match "localhost" string

    # Test full URL matching for localhost with port
    request = MockRequest("http://localhost:9000/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["http://localhost:9000"]
    )
    assert result is request

    # Test port mismatch
    request = MockRequest("http://localhost:8080/test")
    result = ls_utils.filter_request_headers(
        request, allow_hosts=["http://localhost:3000"]
    )
    assert result is None



"""Tests for ExperimentResults edge cases."""

import uuid
from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest

from langsmith import schemas as ls_schemas
from langsmith.evaluation._runner import ExperimentResultRow, ExperimentResults


def _make_result(i: int) -> ExperimentResultRow:
    """Create a minimal ExperimentResultRow for testing."""
    example_id = uuid.uuid4()
    run_id = uuid.uuid4()
    return ExperimentResultRow(
        run=ls_schemas.Run(
            id=run_id,
            name=f"run-{i}",
            run_type="chain",
            inputs={"in": i},
            outputs={"out": i},
            start_time=datetime.now(tz=timezone.utc),
            trace_id=run_id,
            dotted_order=f"20210901T000000000000{run_id}",
        ),
        example=ls_schemas.Example(
            id=example_id,
            dataset_id=uuid.uuid4(),
            inputs={"in": i},
            created_at=datetime.now(tz=timezone.utc),
        ),
        evaluation_results={"results": []},
    )


def _make_manager(
    results: list[ExperimentResultRow],
    summary: dict | None = None,
    error_after: int | None = None,
    summary_error: BaseException | None = None,
):
    """Build a mock _ExperimentManager.

    Args:
        results: Results to yield from get_results().
        summary: Summary scores to return. Defaults to empty.
        error_after: If set, raise RuntimeError after yielding this many results.
        summary_error: If set, raise this from get_summary_scores().
    """
    manager = MagicMock()
    manager.experiment_name = "test-experiment"

    def _get_results():
        for i, r in enumerate(results):
            if error_after is not None and i >= error_after:
                raise RuntimeError(f"producer failed at index {i}")
            yield r

    manager.get_results = _get_results

    if summary_error is not None:
        manager.get_summary_scores = MagicMock(side_effect=summary_error)
    else:
        manager.get_summary_scores = MagicMock(return_value=summary or {"results": []})

    return manager


def test_iterates_all_results():
    """All results from the producer are yielded to the consumer."""
    items = [_make_result(i) for i in range(5)]
    manager = _make_manager(items)

    results = ExperimentResults(manager, blocking=False)
    collected = list(results)

    assert len(collected) == 5
    for i, r in enumerate(collected):
        assert r["run"].name == f"run-{i}"


def test_empty_results():
    """Iterating with zero results immediately stops."""
    manager = _make_manager([])
    results = ExperimentResults(manager, blocking=False)

    collected = list(results)
    assert collected == []


def test_exception_propagates_from_producer():
    """If get_results() raises, __iter__ surfaces the real exception."""
    items = [_make_result(i) for i in range(5)]
    manager = _make_manager(items, error_after=3)

    results = ExperimentResults(manager, blocking=False)
    collected = []
    with pytest.raises(RuntimeError, match="producer failed at index 3"):
        for r in results:
            collected.append(r)

    # The 3 results yielded before the error should have been delivered.
    assert len(collected) == 3


def test_summary_error_propagates():
    """If get_summary_scores() raises, __iter__ surfaces the exception."""
    items = [_make_result(i) for i in range(2)]
    manager = _make_manager(items, summary_error=RuntimeError("summary failed"))

    results = ExperimentResults(manager, blocking=False)
    collected = []
    with pytest.raises(RuntimeError, match="summary failed"):
        for r in results:
            collected.append(r)

    # All items were yielded before the summary error.
    assert len(collected) == 2


def test_wait_raises_on_producer_error():
    """.wait() re-raises the producer exception."""
    items = [_make_result(i) for i in range(3)]
    manager = _make_manager(items, error_after=1)

    results = ExperimentResults(manager, blocking=False)
    with pytest.raises(RuntimeError, match="producer failed at index 1"):
        results.wait()


def test_wait_completes_on_success():
    """.wait() returns cleanly when the producer succeeds."""
    items = [_make_result(i) for i in range(3)]
    manager = _make_manager(items)

    results = ExperimentResults(manager, blocking=False)
    results.wait()
    assert len(results) == 3


def test_len_reflects_buffered_results():
    """len() reflects how many results the producer has buffered so far."""
    items = [_make_result(i) for i in range(4)]
    manager = _make_manager(items)

    results = ExperimentResults(manager, blocking=False)
    results.wait()
    assert len(results) == 4


def test_blocking_mode_propagates_error():
    """In blocking mode, the error propagates during __init__."""
    items = [_make_result(i) for i in range(3)]
    manager = _make_manager(items, error_after=1)

    # In blocking mode, _process_data runs in __init__ — error is stored,
    # then surfaced when iterating.
    results = ExperimentResults(manager, blocking=True)
    collected = []
    with pytest.raises(RuntimeError, match="producer failed at index 1"):
        for r in results:
            collected.append(r)

    assert len(collected) == 1



import asyncio
import json
from unittest import mock
from uuid import uuid4

import pytest
from agents import Agent, Runner, function_tool, set_trace_processors

import langsmith
from langsmith.wrappers import OpenAIAgentsTracingProcessor, wrap_openai
from tests.integration_tests.test_client import safe_delete_dataset

LS_TEST_CLIENT_INFO = {
    "batch_ingest_config": {
        "use_multipart_endpoint": False,
        "scale_up_qsize_trigger": 1000,
        "scale_up_nthreads_limit": 16,
        "scale_down_nempty_trigger": 4,
        "size_limit": 100,
        "size_limit_bytes": 20971520,
    },
}


def _collect_trace_requests(mock_session: mock.MagicMock):
    """Collect and parse trace data from mock session requests."""
    collected_requests = {}
    mock_requests = mock_session.request.call_args_list

    for call in mock_requests:
        if json_bytes := call.kwargs.get("data"):
            json_str = json_bytes.decode("utf-8")
            collected_requests.update(json.loads(json_str))

    return collected_requests


@pytest.mark.asyncio
async def test_openai_agents_tracing_processor():
    """Test that OpenAIAgentsTracingProcessor correctly traces agent runs."""
    import openai

    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session, info=LS_TEST_CLIENT_INFO)

    processor = OpenAIAgentsTracingProcessor(client=client)
    set_trace_processors([processor])

    agent = Agent(
        name="Captain Obvious",
        instructions="You are Captain Obvious, the world's"
        + " most literal technical support agent.",
    )

    question = (
        "Why is my code failing when I try to divide by zero?"
        " I keep getting this error message."
    )
    try:
        result = await Runner.run(agent, question)
    except openai.APIConnectionError as e:
        pytest.skip(reason="Openai is having issues" + str(e))

    # Verify we got a result
    assert result is not None
    assert hasattr(result, "final_output")
    assert result.final_output is not None

    # Give the background thread time to process traces
    await asyncio.sleep(0.5)

    # Verify that trace calls were made to the mock session
    assert mock_session.request.call_count > 0

    # Collect and verify trace data was sent
    collected_requests = _collect_trace_requests(mock_session)

    # Verify we have trace events
    all_events = [
        *collected_requests.get("post", []),
        *collected_requests.get("patch", []),
    ]

    assert len(all_events) > 0, "No trace events were recorded"

    # Verify that we have both start and end events (at least one should have end_time)
    has_start_event = any(event.get("start_time") for event in all_events)
    has_end_event = any(event.get("end_time") for event in all_events)

    assert has_start_event, "No trace start events found"
    assert has_end_event, "No trace end events found"

    # Verify trace contains expected agent information
    agent_runs = [
        event for event in all_events if event.get("name") == "Agent workflow"
    ]
    assert len(agent_runs) > 0, "No agent workflow runs found in trace"

    agent_run = agent_runs[0]
    metadata = (agent_run.get("extra") or {}).get("metadata") or {}
    assert metadata.get("ls_integration") == "openai-agents-sdk", agent_run
    assert "ls_integration_version" in metadata, agent_run


@pytest.mark.xfail(reason="Flaky test - may fail intermittently")
async def test_openai_agents_with_evaluate():
    client = langsmith.Client()

    processor = OpenAIAgentsTracingProcessor(client=client)
    set_trace_processors([processor])

    agent = Agent(
        name="Captain Obvious",
        instructions="You are Captain Obvious, the world's"
        + " most literal technical support agent.",
    )

    question = (
        "Why is my code failing when I try to divide by zero?"
        " I keep getting this error message."
    )

    # Example: Cat image validation
    question = "What type of cat is shown in this image?"
    provided_answer = "This is a tuxedo cat with black and white fur pattern."
    image_url = "https://upload.wikimedia.org/wikipedia/commons/thumb/b/b7/George%2C_a_perfect_example_of_a_tuxedo_cat.jpg/1250px-George%2C_a_perfect_example_of_a_tuxedo_cat.jpg"

    dataset_name = "OpenAI Agent Testing" + str(uuid4().hex[:4])

    if client.has_dataset(dataset_name=dataset_name):
        safe_delete_dataset(client, dataset_name=dataset_name)
    try:
        dataset = client.create_dataset(dataset_name=dataset_name)
        client.create_examples(
            inputs=[
                {
                    "question": question,
                    "answer": provided_answer,
                    "image_url": image_url,
                }
            ],
            outputs=[{"is_correct": True}],
            dataset_id=dataset.id,
        )

        async def run_agent(inputs: dict):
            """Run agent and track the final response."""
            await Runner.run(agent, question)
            return {"result": "foo"}

        async def alignment(outputs: dict, reference_outputs: dict) -> bool:
            """Check if the agent chose the correct route."""
            return True

        experiment = await client.aevaluate(
            run_agent,
            data=dataset_name,
            evaluators=[alignment],
            experiment_prefix="agent-gpt-5-mini",
            max_concurrency=2,
            blocking=True,
        )
        await asyncio.sleep(5)
        experiment_results = client.get_experiment_results(
            name=experiment.experiment_name
        )
        assert experiment_results["stats"].run_count == 1
        assert (
            experiment_results["stats"].feedback_stats.get("alignment", {}).get("n")
            == 1
        )
        assert (
            experiment_results["stats"].feedback_stats.get("alignment", {}).get("avg")
            == 1
        )
        examples = list(experiment_results["examples_with_runs"])
        assert len(examples) == 1
        run = client.read_run(examples[0].runs[0].id, load_child_runs=True)
        assert len(run.child_runs) == 1
        assert run.child_runs[0].name == "Agent workflow"
        assert len(run.child_runs[0].child_runs) == 1
        assert run.child_runs[0].child_runs[0].name == "Captain Obvious"
    finally:
        safe_delete_dataset(client, dataset_name=dataset_name)


@pytest.mark.asyncio
async def test_wrap_openai_nests_under_agent_trace():
    """Test that wrap_openai calls from function tools nest under agent traces."""
    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session, info=LS_TEST_CLIENT_INFO)

    try:
        import openai

        openai_client = wrap_openai(openai.OpenAI(api_key="test-key"))
    except ImportError:
        pytest.skip("openai package not installed")

    # Create a function tool that uses wrap_openai
    @function_tool
    def call_openai_helper(query: str) -> str:
        """A helper function that calls OpenAI using wrap_openai."""
        try:
            response = openai_client.chat.completions.create(
                model="gpt-4o-mini",
                messages=[{"role": "user", "content": query}],
            )
            return response.choices[0].message.content or "No response"
        except Exception as e:
            return f"Error: {e}"

    # Set up the tracing processor
    processor = OpenAIAgentsTracingProcessor(client=client)
    set_trace_processors([processor])

    agent = Agent(
        name="Test Agent",
        instructions="You are a helpful assistant.",
        tools=[call_openai_helper],
    )

    # Run the agent
    try:
        await Runner.run(
            agent, "Use the call_openai_helper tool to answer: What is 2+2?"
        )
    except Exception:
        pass

    await asyncio.sleep(0.5)

    collected_requests = _collect_trace_requests(mock_session)

    all_events = [
        *collected_requests.get("post", []),
        *collected_requests.get("patch", []),
    ]

    assert len(all_events) > 0, "No trace events were recorded"

    runs_by_id = {event["id"]: event for event in all_events if "id" in event}

    agent_runs = [
        event for event in all_events if event.get("name") == "Agent workflow"
    ]
    assert len(agent_runs) > 0, "No agent workflow runs found"

    chat_runs = [event for event in all_events if "ChatOpenAI" in event.get("name", "")]

    if chat_runs:
        for chat_run in chat_runs:
            assert chat_run.get("parent_run_id") is not None, (
                "ChatOpenAI run should have a parent"
            )

            current_id = chat_run.get("parent_run_id")
            found_agent = False
            depth = 0
            max_depth = 10

            while current_id and depth < max_depth:
                parent_run = runs_by_id.get(current_id)
                if not parent_run:
                    break

                if parent_run.get("name") == "Agent workflow":
                    found_agent = True
                    break

                current_id = parent_run.get("parent_run_id")
                depth += 1

            assert found_agent, "ChatOpenAI run should be nested under Agent workflow"

    trace_ids = {event.get("trace_id") for event in all_events if "trace_id" in event}
    assert len(trace_ids) <= 2, (
        f"Expected at most 2 trace_ids (one for the trace, possibly one for root), "
        f"but found {len(trace_ids)}: {trace_ids}"
    )

    # Verify invocation_params with tools are present in trace events
    events_with_tools = [
        event
        for event in all_events
        if event.get("extra", {}).get("invocation_params", {}).get("tools")
    ]
    assert len(events_with_tools) > 0, (
        "No trace events found with invocation_params.tools - "
        "tools should be captured in invocation_params"
    )


@pytest.mark.asyncio
async def test_traceable_decorator_nests_under_agent_trace():
    """Test that @traceable decorated functions nest under agent traces."""
    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session, info=LS_TEST_CLIENT_INFO)

    @langsmith.traceable
    def helper_function(x: int, y: int) -> int:
        return x + y

    @function_tool
    def calculate_sum(a: int, b: int) -> str:
        result = helper_function(a, b)
        return f"The sum is {result}"

    processor = OpenAIAgentsTracingProcessor(client=client)
    set_trace_processors([processor])

    agent = Agent(
        name="Calculator Agent",
        instructions="You are a calculator.",
        tools=[calculate_sum],
    )

    try:
        await Runner.run(agent, "Calculate 5 + 3")
    except Exception:
        pass

    await asyncio.sleep(0.5)

    collected_requests = _collect_trace_requests(mock_session)

    all_events = [
        *collected_requests.get("post", []),
        *collected_requests.get("patch", []),
    ]

    assert len(all_events) > 0, "No trace events were recorded"

    runs_by_id = {event["id"]: event for event in all_events if "id" in event}

    agent_runs = [
        event for event in all_events if event.get("name") == "Agent workflow"
    ]
    assert len(agent_runs) > 0, "No agent workflow runs found"

    helper_runs = [
        event for event in all_events if event.get("name") == "helper_function"
    ]

    if helper_runs:
        for helper_run in helper_runs:
            assert helper_run.get("parent_run_id") is not None, (
                "helper_function run should have a parent"
            )

            current_id = helper_run.get("parent_run_id")
            found_agent = False
            depth = 0
            max_depth = 10

            while current_id and depth < max_depth:
                parent_run = runs_by_id.get(current_id)
                if not parent_run:
                    break

                if parent_run.get("name") == "Agent workflow":
                    found_agent = True
                    break

                current_id = parent_run.get("parent_run_id")
                depth += 1

            assert found_agent, (
                "helper_function run should be nested under Agent workflow"
            )



# mypy: disable-error-code="annotation-unchecked"
import json
import re
import uuid
from typing import List, Union, cast
from unittest.mock import MagicMock
from uuid import uuid4

from pydantic import BaseModel

from langsmith import Client, traceable, tracing_context
from langsmith.anonymizer import RuleNodeProcessor, StringNodeRule, create_anonymizer

EMAIL_REGEX = re.compile(r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}")
UUID_REGEX = re.compile(
    r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
)


def test_replacer_function():
    def replacer(text: str, _: List[Union[str, int]]):
        text = EMAIL_REGEX.sub("[email address]", text)
        text = UUID_REGEX.sub("[uuid]", text)
        return text

    assert create_anonymizer(replacer)(
        {
            "message": "Hello, this is my email: hello@example.com",
            "metadata": str(uuid4()),
        }
    ) == {
        "message": "Hello, this is my email: [email address]",
        "metadata": "[uuid]",
    }

    assert create_anonymizer(replacer)(["human", "hello@example.com"]) == [
        "human",
        "[email address]",
    ]
    assert create_anonymizer(replacer)("hello@example.com") == "[email address]"


def test_replacer_lambda():
    assert create_anonymizer(lambda text: EMAIL_REGEX.sub("[email address]", text))(
        {
            "message": "Hello, this is my email: hello@example.com",
        }
    ) == {
        "message": "Hello, this is my email: [email address]",
    }


def test_replacer_declared():
    replacers = [
        StringNodeRule(pattern=EMAIL_REGEX, replace="[email address]"),
        StringNodeRule(pattern=UUID_REGEX, replace="[uuid]"),
    ]

    assert create_anonymizer(replacers)(
        {
            "message": "Hello, this is my email: hello@example.com",
            "metadata": str(uuid4()),
        }
    ) == {
        "message": "Hello, this is my email: [email address]",
        "metadata": "[uuid]",
    }

    assert create_anonymizer(replacers)(["human", "hello@example.com"]) == [
        "human",
        "[email address]",
    ]

    assert create_anonymizer(replacers)("hello@example.com") == "[email address]"


def test_replacer_declared_in_traceable():
    replacers = [
        StringNodeRule(pattern=EMAIL_REGEX, replace="[email address]"),
        StringNodeRule(pattern=UUID_REGEX, replace="[uuid]"),
    ]
    anonymizer = create_anonymizer(replacers)
    mock_client = Client(
        session=MagicMock(),
        auto_batch_tracing=False,
        anonymizer=anonymizer,
        api_url="http://localhost:1984",
        api_key="123",
    )

    user_email = "my-test@langchain.ai"
    user_id = "4ae21a90-d43b-4017-bb21-4fd9add235ff"

    class MyOutput(BaseModel):
        user_email: str
        user_id: uuid.UUID
        body: str

    class MyInput(BaseModel):
        from_email: str

    @traceable(client=mock_client)
    def my_func(body: str, from_: MyInput) -> MyOutput:
        return MyOutput(user_email=user_email, user_id=user_id, body=body)

    body_ = "Hello from Pluto"
    with tracing_context(enabled=True):
        res = my_func(body_, from_=MyInput(from_email="my-from-test@langchain.ai"))
    expected = MyOutput(user_email=user_email, user_id=uuid.UUID(user_id), body=body_)
    assert res == expected
    # get posts
    posts = [
        json.loads(call[2]["data"])
        for call in mock_client.session.request.mock_calls
        if call.args and call.args[1].endswith("runs")
    ]

    patches = [
        json.loads(call[2]["data"])
        for call in mock_client.session.request.mock_calls
        if call.args
        and cast(str, call.args[0]).lower() == "patch"
        and "/runs" in call.args[1]
    ]

    expected_inputs = {"from_": {"from_email": "[email address]"}, "body": body_}
    expected_outputs = {
        "user_email": "[email address]",
        "user_id": "[uuid]",
        "body": body_,
    }

    assert len(posts) == 1
    posted_data = posts[0]
    assert posted_data["inputs"] == expected_inputs
    assert len(patches) == 1
    patched_data = patches[0]
    if "inputs" in patched_data:
        assert patched_data["inputs"] == expected_inputs
    assert patched_data["outputs"] == expected_outputs


def test_rule_node_processor_scrub_sensitive_info():
    rules = [
        StringNodeRule(pattern=re.compile(r"\b\d{3}-\d{2}-\d{4}\b"), replace="[ssn]"),
        StringNodeRule(
            pattern=re.compile(r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}"),
            replace="[email]",
        ),
        StringNodeRule(
            pattern=re.compile(r"\b\d{3}[-.\s]?\d{3}[-.\s]?\d{4}\b"), replace="[phone]"
        ),
    ]
    processor = RuleNodeProcessor(rules)

    nodes = [
        {"value": "My SSN is 123-45-6789.", "path": ["field1"]},
        {"value": "Contact me at john.doe@example.com.", "path": ["field2"]},
        {"value": "Call me on 123-456-7890.", "path": ["field3"]},
    ]

    expected = [
        {"value": "My SSN is [ssn].", "path": ["field1"]},
        {"value": "Contact me at [email].", "path": ["field2"]},
        {"value": "Call me on [phone].", "path": ["field3"]},
    ]

    result = processor.mask_nodes(nodes)

    assert result == expected


def test_rule_node_processor_default_replace():
    rules = [
        StringNodeRule(pattern=re.compile(r"sensitive")),
    ]
    processor = RuleNodeProcessor(rules)

    nodes = [
        {"value": "This contains sensitive data", "path": ["field1"]},
    ]

    expected = [
        {"value": "This contains [redacted] data", "path": ["field1"]},
    ]

    result = processor.mask_nodes(nodes)
    assert result == expected



# mypy: disable-error-code="attr-defined, union-attr, arg-type, call-overload"
from __future__ import annotations

import json
import time
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from langsmith import Client
from langsmith.wrappers import wrap_gemini
from langsmith.wrappers._gemini import _process_generate_content_response
from tests.unit_tests.test_run_helpers import _get_calls

if TYPE_CHECKING:
    from google import genai

model_name = "gemini-2.0-flash-001"
prompt = "Say 'foo'"


LS_TEST_CLIENT_INFO = {
    "batch_ingest_config": {
        "use_multipart_endpoint": False,
        "scale_up_qsize_trigger": 1000,
        "scale_up_nthreads_limit": 16,
        "scale_down_nempty_trigger": 4,
        "size_limit": 100,
        "size_limit_bytes": 20971520,
    },
}


@pytest.fixture
def original_client() -> genai.Client:
    """Create an original Gemini client."""
    from google import genai

    return genai.Client()


@pytest.fixture
def mock_ls_client() -> Client:
    """Create a mock LangSmith client."""
    mock_session = mock.MagicMock()
    return Client(session=mock_session, info=LS_TEST_CLIENT_INFO)


@pytest.fixture
def patched_client(mock_ls_client: Client) -> genai.Client:
    """Create a patched Gemini client with tracing."""
    from google import genai

    return wrap_gemini(genai.Client(), tracing_extra={"client": mock_ls_client})


def test_generate_content_sync(
    original_client: genai.Client,
    patched_client: genai.Client,
    mock_ls_client: Client,
):
    """Test non-streaming generate_content."""
    original = original_client.models.generate_content(
        model=model_name,
        contents=prompt,
    )
    patched = patched_client.models.generate_content(
        model=model_name,
        contents=prompt,
    )

    # Verify responses are similar
    assert isinstance(patched, type(original))
    assert original.text == patched.text

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(mock_ls_client, minimum=1)
    assert calls
    datas = [json.loads(call.kwargs["data"]) for call in calls]

    # Check that outputs were captured
    outputs = None
    for data in datas:
        if data.get("post"):
            if outputs := data["post"][0].get("outputs"):
                break
        if data.get("patch"):
            outputs = data["patch"][0].get("outputs")
            break
    assert outputs


def test_generate_content_stream_sync(
    original_client: genai.Client,
    patched_client: genai.Client,
    mock_ls_client: Client,
):
    """Test streaming generate_content_stream."""
    original_chunks = []
    for chunk in original_client.models.generate_content_stream(
        model=model_name,
        contents=prompt,
    ):
        if chunk.text:
            original_chunks.append(chunk.text)

    patched_chunks = []
    for chunk in patched_client.models.generate_content_stream(
        model=model_name,
        contents=prompt,
    ):
        if chunk.text:
            patched_chunks.append(chunk.text)

    # Verify responses are similar
    assert len(original_chunks) == len(patched_chunks)
    assert "".join(original_chunks) == "".join(patched_chunks)

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(mock_ls_client, minimum=1)
    assert calls
    datas = [json.loads(call.kwargs["data"]) for call in calls]

    # Check that outputs were captured
    outputs = None
    for data in datas:
        if data.get("post"):
            if outputs := data["post"][0].get("outputs"):
                break
        if data.get("patch"):
            outputs = data["patch"][0].get("outputs")
            break
    assert outputs


@pytest.mark.asyncio
async def test_generate_content_async():
    """Test async generate_content."""
    from google import genai

    mock_session = mock.MagicMock()
    client = Client(session=mock_session)

    # Create async client
    original_client = genai.Client()
    patched_client = wrap_gemini(genai.Client(), tracing_extra={"client": client})

    # Test non-streaming async
    original = await original_client.aio.models.generate_content(
        model=model_name,
        contents=prompt,
    )
    patched = await patched_client.aio.models.generate_content(
        model=model_name,
        contents=prompt,
    )

    # Verify responses are similar
    assert isinstance(patched, type(original))
    assert original.text == patched.text

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(client, minimum=1)
    assert calls


@pytest.mark.asyncio
async def test_generate_content_stream_async():
    """Test async streaming generate_content_stream."""
    from google import genai

    mock_session = mock.MagicMock()
    client = Client(session=mock_session)

    # Create async client
    original_client = genai.Client()
    patched_client = wrap_gemini(genai.Client(), tracing_extra={"client": client})

    # Test streaming async
    original_chunks = []
    async for chunk in await original_client.aio.models.generate_content_stream(
        model=model_name,
        contents=prompt,
    ):
        if chunk.text:
            original_chunks.append(chunk.text)

    patched_chunks = []
    async for chunk in await patched_client.aio.models.generate_content_stream(
        model=model_name,
        contents=prompt,
    ):
        if chunk.text:
            patched_chunks.append(chunk.text)

    # Verify responses are similar
    assert len(original_chunks) == len(patched_chunks)
    assert "".join(original_chunks) == "".join(patched_chunks)

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(client, minimum=1)
    assert calls


def test_custom_config(patched_client: genai.Client, mock_ls_client: Client):
    """Test with custom configuration."""
    from google.genai import types

    config = types.GenerateContentConfig(
        temperature=0.5,
        max_output_tokens=100,
        stop_sequences=["END"],
    )

    response = patched_client.models.generate_content(
        model=model_name,
        contents=prompt,
        config=config,
    )

    assert response.text

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(mock_ls_client, minimum=1)
    assert calls


def test_finish_reason(patched_client: genai.Client, mock_ls_client: Client):
    """Test finish reason."""
    from google.genai.types import FinishReason

    response = patched_client.models.generate_content(
        model=model_name,
        contents=prompt,
    )
    assert response.candidates[0].finish_reason == FinishReason.STOP


def test_multimodal_image_generation(
    patched_client: genai.Client, mock_ls_client: Client
):
    """Test multimodal response with both text and image output."""
    multimodal_prompt = (
        "Create a picture of a nano banana dish in a fancy restaurant "
        "with a Gemini theme"
    )

    response = patched_client.models.generate_content(
        model="gemini-2.5-flash-image",
        contents=[multimodal_prompt],
    )

    # Verify response contains expected parts
    assert response.candidates
    assert response.candidates[0].content.parts

    # Check that we got both text and image parts
    has_text = False
    has_image = False

    for part in response.candidates[0].content.parts:
        if part.text is not None:
            has_text = True
        elif part.inline_data is not None:
            has_image = True
            # Verify inline_data has the expected structure
            assert part.inline_data.data
            assert part.inline_data.mime_type

    # For this specific prompt, we expect both text and image
    assert has_text or has_image, "Response should contain text or image"

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(mock_ls_client, minimum=1)
    assert calls

    # Verify that multimodal content was properly traced
    datas = [json.loads(call.kwargs["data"]) for call in calls]
    outputs = None
    for data in datas:
        if data.get("post"):
            if outputs := data["post"][0].get("outputs"):
                break
        if data.get("patch"):
            outputs = data["patch"][0].get("outputs")
            break

    assert outputs
    # Output should have content (our chat-like format)
    assert "content" in outputs or "role" in outputs


def test_multimodal_input_with_text_only(
    patched_client: genai.Client, mock_ls_client: Client
):
    """Test that multimodal input structure (list of contents) works correctly."""
    # Use list format for contents (multimodal structure) but only with text
    multimodal_input = [{"role": "user", "parts": [{"text": "What is AI in 5 words?"}]}]

    response = patched_client.models.generate_content(
        model=model_name,
        contents=multimodal_input,
    )

    assert response.text
    assert len(response.text) > 0

    # Give the thread a chance to post traces
    time.sleep(0.1)

    # Verify tracing calls were made
    calls = _get_calls(mock_ls_client, minimum=1)
    assert calls

    # Verify input was normalized to messages format
    datas = [json.loads(call.kwargs["data"]) for call in calls]
    for data in datas:
        if data.get("post"):
            inputs = data["post"][0].get("inputs")
            if inputs:
                # Our _process_gemini_inputs should convert to messages format
                assert "messages" in inputs
                assert isinstance(inputs["messages"], list)
                assert len(inputs["messages"]) > 0
                assert inputs["messages"][0]["role"] == "user"
                break


def test_function_calling(patched_client: genai.Client, mock_ls_client: Client):
    """Test function calling with manual function declaration."""
    from google.genai import types

    # Manually define the function declaration
    get_weather_func = types.FunctionDeclaration(
        name="get_current_weather",
        description="Get the current weather in a given location",
        parameters={
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                }
            },
            "required": ["location"],
        },
    )

    response = patched_client.models.generate_content(
        model=model_name,
        contents="What is the weather like in Boston?",
        config=types.GenerateContentConfig(
            tools=[types.Tool(function_declarations=[get_weather_func])],
        ),
    )

    # Test the processing function
    result = _process_generate_content_response(response)

    # Check that function call was made
    assert response.candidates
    assert response.candidates[0].content.parts

    # Verify the function call is present in result
    assert "tool_calls" in result, f"Expected tool_calls, got: {result.keys()}"
    assert len(result["tool_calls"]) >= 1

    tool_call = result["tool_calls"][0]
    assert tool_call["type"] == "function"
    assert tool_call["function"]["name"] == "get_current_weather"
    assert "location" in json.loads(tool_call["function"]["arguments"])



import pytest

from langsmith.env import __all__ as env_all
from langsmith.env import get_git_info

_EXPECTED = [
    "get_docker_compose_command",
    "get_docker_compose_version",
    "get_docker_environment",
    "get_docker_version",
    "get_langchain_env_var_metadata",
    "get_langchain_env_vars",
    "get_langchain_environment",
    "get_release_shas",
    "get_runtime_and_metrics",
    "get_runtime_environment",
    "get_system_metrics",
    "get_git_info",
]


def test_public_api() -> None:
    assert env_all == _EXPECTED


def test_git_info() -> None:
    git_info = get_git_info()
    try:
        assert git_info is not None
        assert git_info["commit"] is not None
        assert git_info["remote_url"] is not None
        assert "langsmith-sdk" in git_info["remote_url"]
    except AssertionError:
        pytest.skip("Git information is not available, skipping test.")



import asyncio

import httpx
import pytest
from uvicorn import Config, Server

from langsmith import traceable
from langsmith.run_helpers import get_current_run_tree
from tests.integration_tests.fake_server import fake_app


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module")
async def fake_server():
    config = Config(app=fake_app, loop="asyncio", port=8000, log_level="info")
    server = Server(config=config)

    asyncio.create_task(server.serve())
    await asyncio.sleep(0.1)

    yield
    try:
        await server.shutdown()
    except RuntimeError:
        pass


@traceable
async def the_parent_function():
    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=fake_app),
        base_url="http://localhost:8000",
    ) as http_client:
        headers = {}
        if span := get_current_run_tree():
            headers.update(span.to_headers())
        response = await http_client.post("/fake-route", headers=headers)
        assert response.status_code == 200
        return response.json()


@traceable
async def the_root_function(foo: str):
    return await the_parent_function()


@pytest.mark.asyncio
async def test_tracing_fake_server(fake_server):
    result = await the_root_function(
        "test input",
        langsmith_extra={
            "metadata": {"some-cool-value": 42},
            "tags": ["did-propagate"],
            "project_name": "distributed-tracing",
        },
    )
    assert result["message"] == "Fake route response"



import uuid
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import Any, Callable, Dict, Optional

import numpy as np
from pydantic import BaseModel, Field
from pydantic.v1 import BaseModel as BaseModelV1
from pydantic.v1 import Field as FieldV1


def _default():
    return {
        "some_val": "😈",
        "uuid_val": uuid.uuid4(),
        "datetime_val": datetime.now(),
        "list_val": [238928376271863487] * 5,
        "decimal_val": Decimal("3.14"),
        "set_val": {1, 2, 3},
        "tuple_val": (4, 5, 6),
        "bytes_val": b"hello world",
        "arr": np.random.random(10),
    }


@dataclass
class DeeplyNested:
    """An object."""

    vals: Dict[str, Any] = field(default_factory=_default)


class DeeplyNestedModel(BaseModel):
    vals: Dict[str, Any] = Field(default_factory=_default)


class DeeplyNestedModelV1(BaseModelV1):
    vals: Dict[str, Any] = FieldV1(default_factory=_default)


def create_nested_instance(
    depth: int = 5,
    width: int = 5,
    branch_constructor: Optional[Callable] = DeeplyNested,
    leaf_constructor: Optional[Callable] = None,
) -> DeeplyNested:
    top_level = DeeplyNested()
    current_level = top_level
    root_constructor = leaf_constructor or DeeplyNested
    for i in range(depth):
        for j in range(width):
            key = f"key_{i}_{j}"
            if i < depth - 1:
                value = branch_constructor()
                current_level.vals[key] = value
                if j == 0:
                    next_level = value
            else:
                current_level.vals[key] = root_constructor()

        if i < depth - 1:
            current_level = next_level
    return top_level


if __name__ == "__main__":
    import time

    from langsmith.client import _dumps_json

    class MyClass:
        def __init__(self):
            self.vals = {}

    def run():
        res = create_nested_instance(200, 150, leaf_constructor=MyClass)
        start_time = time.time()
        res = _dumps_json({"input": res})
        end_time = time.time()
        print(f"Size: {len(res) / 1024:.2f} KB")
        print(f"Time taken: {end_time - start_time:.2f} seconds")

    run()



"""Common test utilities for unit tests."""

from __future__ import annotations

import json
import re
from typing import Any, Dict, List


def parse_multipart_data(data: bytes) -> Dict[str, Any]:
    """Parse multipart form data into a batch-like dict structure.

    Returns a dict with "post" and "patch" keys containing lists of runs,
    similar to the batch endpoint format.
    """
    if isinstance(data, str):
        data = data.encode("utf-8")

    result: Dict[str, List[Dict[str, Any]]] = {"post": [], "patch": []}

    # Extract boundary from content type or detect from data
    boundary_match = re.search(rb"--([^\r\n]+)", data)
    if not boundary_match:
        return result

    boundary = boundary_match.group(1)

    # Split by boundary
    parts = data.split(b"--" + boundary)

    # Track runs by ID to merge main payload with field parts
    runs: Dict[str, Dict[str, Any]] = {}
    run_methods: Dict[str, str] = {}  # track whether post or patch

    for part in parts:
        if not part or part.strip() in (b"", b"--", b"--\r\n"):
            continue

        # Parse headers and body
        try:
            if b"\r\n\r\n" in part:
                headers_section, body = part.split(b"\r\n\r\n", 1)
            else:
                continue

            # Remove trailing boundary markers
            body = body.rstrip(b"\r\n")

            # Extract name from Content-Disposition
            name_match = re.search(rb'name="([^"]+)"', headers_section)
            if not name_match:
                continue

            name = name_match.group(1).decode("utf-8")

            # Parse the name: format is "method.run_id" or "method.run_id.field"
            name_parts = name.split(".")

            if len(name_parts) >= 2:
                method = name_parts[0]  # "post" or "patch"
                run_id = name_parts[1]

                if method not in ("post", "patch"):
                    continue

                # Initialize run dict if needed
                if run_id not in runs:
                    runs[run_id] = {}
                    run_methods[run_id] = method

                if len(name_parts) == 2:
                    # Main payload: method.run_id
                    try:
                        run_data = json.loads(body)
                        runs[run_id].update(run_data)
                    except json.JSONDecodeError:
                        pass
                elif len(name_parts) == 3:
                    # Field payload: method.run_id.field
                    field = name_parts[2]
                    try:
                        field_data = json.loads(body)
                        runs[run_id][field] = field_data
                    except json.JSONDecodeError:
                        pass

        except Exception:
            continue

    # Organize into post/patch lists
    for run_id, run_data in runs.items():
        method = run_methods.get(run_id, "post")
        result[method].append(run_data)

    return result


def parse_request_data(data: bytes, content_type: str = "") -> Dict[str, Any]:
    """Parse request data, handling both JSON and multipart formats.

    This function automatically detects the format based on content type
    or data structure.
    """
    if isinstance(data, str):
        data = data.encode("utf-8")

    # Try JSON first
    if "multipart" not in content_type.lower():
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            pass

    # Try multipart
    if data.startswith(b"--"):
        return parse_multipart_data(data)

    # Fallback to JSON
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        return {"post": [], "patch": []}



import importlib
import os
import random
import time

import pytest

from langsmith import testing as t
from langsmith import traceable

pytestmark = pytest.mark.skipif(
    not os.getenv("LANGSMITH_TRACING"),
    reason="LANGSMITH_TRACING environment variable not set",
)


@pytest.mark.langsmith
@pytest.mark.parametrize("c", list(range(10)))
async def test_addition_single(c):
    x = 3
    y = 4
    t.log_inputs({"x": x, "y": y, "c": c})

    expected = 7 + c
    t.log_reference_outputs({"sum": expected})

    actual = x + y + c
    t.log_outputs({"sum": actual})

    t.log_feedback(key="foo", score=1)

    assert actual == expected


async def my_app():
    return "hello"


@pytest.mark.langsmith
async def test_openai_says_hello():
    # Traced code will be included in the test case
    text = "Say hello!"
    response = await my_app()
    t.log_inputs({"text": text})
    t.log_outputs({"response": response})
    t.log_reference_outputs({"response": "hello!"})

    # Use this context manager to trace any steps used for generating evaluation
    # feedback separately from the main application logic
    with t.trace_feedback():
        grade = 1 if "hello" in response else 0
        t.log_feedback(key="llm_judge", score=grade)

    assert "hello" in response.lower()


@pytest.mark.langsmith
async def test_composite_evaluator():
    # Traced code will be included in the test case
    text = "Say hello!"
    response = await my_app()
    t.log_inputs({"text": text})
    t.log_outputs({"response": response})
    t.log_reference_outputs({"response": "hello!"})

    @traceable
    def my_composite_evaluator(response):
        with t.trace_feedback():
            grade = 1 if "hello" in response else 0
            t.log_feedback(key="composite_judge", score=grade)
            return grade

    my_composite_evaluator(response)

    assert "hello" in response.lower()


@pytest.mark.xfail(reason="Test failure output case")
@pytest.mark.langsmith(output_keys=["expected"])
@pytest.mark.parametrize(
    "a, b, expected",
    [
        (1, 2, 3),
        (3, 4, 7),
    ],
)
async def test_addition_parametrized(a: int, b: int, expected: int):
    t.log_outputs({"sum": a + b})
    assert a + b != expected


@pytest.mark.langsmith
@pytest.mark.parametrize("a,b", [[i, i] for i in range(20)])
def test_param(a, b):
    t.log_outputs({"sum": a + b})
    t.log_reference_outputs({"sum": a + b})
    assert a + b == a + b


@pytest.fixture
def inputs() -> int:
    return 5


@pytest.fixture
def reference_outputs() -> int:
    return 10


@pytest.mark.langsmith(output_keys=["reference_outputs"])
def test_fixture(inputs: int, reference_outputs: int):
    result = 2 * inputs
    t.log_outputs({"d": result})
    assert result == reference_outputs


@pytest.mark.langsmith
def test_slow_test():
    t.log_inputs({"slow": "I am slow"})
    time.sleep(5)
    t.log_outputs({"slow_result": "I am slow"})
    t.log_reference_outputs({"slow_result": "I am not fast"})


@pytest.mark.skipif(
    not importlib.util.find_spec("langchain_core"),
    reason="langchain-core not installed",
)
@pytest.mark.langsmith
def test_log_langchain_outputs() -> None:
    from langchain_core.messages import AIMessage

    t.log_inputs({"question": "foo"})
    t.log_outputs({"answer": AIMessage("bar")})


@pytest.mark.langsmith(
    metadata={"test_type": "metadata_test", "custom_key": "custom_value"}
)
def test_metadata_parameter():
    """Test that metadata parameter is properly passed to the decorator."""
    x = 5
    y = 10
    t.log_inputs({"x": x, "y": y})

    result = x + y
    t.log_outputs({"sum": result})
    t.log_reference_outputs({"sum": 15})

    assert result == 15


@pytest.mark.langsmith(
    metadata={"test_type": "metadata_test_asyncy", "custom_key": "custom_value_async"}
)
async def test_metadata_parameter_async():
    """Test that metadata parameter is properly passed to the decorator."""
    x = 5
    y = 10
    t.log_inputs({"x": x, "y": y})

    result = x + y
    t.log_outputs({"sum": result})
    t.log_reference_outputs({"sum": 15})

    assert result == 15


@pytest.mark.langsmith(repetitions=2)
def test_repetitions_parameter():
    """Test that repetitions parameter causes test to run multiple times."""
    x = 5
    y = 10
    t.log_inputs({"x": x, "y": y})

    result = x + y
    t.log_outputs({"sum": result, "random": random.random()})
    t.log_reference_outputs({"sum": 15})

    assert result == 15


@pytest.mark.langsmith(split="test_split")
def test_split_parameter():
    """Test that split parameter causes test to run multiple times."""
    x = "test_split"
    t.log_inputs({"split": x})

    t.log_outputs({"foo": "foo"})
    t.log_reference_outputs({"foo": "foo"})

    assert x == "test_split"


@pytest.mark.langsmith(repetitions=3)
async def test_repetitions_parameter_async():
    """Test that repetitions parameter causes async test to run multiple times."""
    x = 5
    y = 10
    t.log_inputs({"x": x, "y": y})

    result = x + y
    t.log_outputs({"sum": result, "random": random.random()})
    t.log_reference_outputs({"sum": 15})

    assert result == 15



"""Tests for RetryTransport and AsyncRetryTransport."""

from __future__ import annotations

from unittest.mock import patch

import httpx
import pytest

from langsmith.sandbox._exceptions import SandboxConnectionError
from langsmith.sandbox._transport import AsyncRetryTransport, RetryTransport


class MockTransport(httpx.BaseTransport):
    """Sync mock transport that returns pre-configured responses in sequence."""

    def __init__(self, responses: list[httpx.Response | Exception]) -> None:
        self._responses = list(responses)
        self._calls: list[httpx.Request] = []

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        self._calls.append(request)
        item = self._responses.pop(0)
        if isinstance(item, Exception):
            raise item
        return item

    @property
    def call_count(self) -> int:
        return len(self._calls)


class MockAsyncTransport(httpx.AsyncBaseTransport):
    """Async mock transport that returns pre-configured responses in sequence."""

    def __init__(self, responses: list[httpx.Response | Exception]) -> None:
        self._responses = list(responses)
        self._calls: list[httpx.Request] = []

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        self._calls.append(request)
        item = self._responses.pop(0)
        if isinstance(item, Exception):
            raise item
        return item

    @property
    def call_count(self) -> int:
        return len(self._calls)


def _make_request() -> httpx.Request:
    return httpx.Request("GET", "http://test-server/volumes")


# ============================================================================
# Sync RetryTransport tests
# ============================================================================


class TestRetryTransport:
    """Tests for the sync RetryTransport."""

    @pytest.mark.parametrize("status", [502, 503, 504])
    @patch("langsmith.sandbox._transport.time.sleep")
    def test_retries_on_server_error_then_succeeds(self, mock_sleep, status):
        mock = MockTransport([httpx.Response(status), httpx.Response(200)])
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2
        assert mock_sleep.call_count == 1

    @pytest.mark.parametrize(
        ("retry_after", "min_sleep"),
        [("2", 2.0), (None, 1.0), ("not-a-number", 1.0)],
        ids=["explicit", "default", "invalid"],
    )
    @patch("langsmith.sandbox._transport.time.sleep")
    def test_retries_on_429(self, mock_sleep, retry_after, min_sleep):
        headers = {"retry-after": retry_after} if retry_after else {}
        mock = MockTransport(
            [
                httpx.Response(429, headers=headers),
                httpx.Response(200),
            ]
        )
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2
        assert mock_sleep.call_args[0][0] >= min_sleep

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_retries_on_connect_error_then_succeeds(self, mock_sleep):
        mock = MockTransport(
            [
                httpx.ConnectError("Connection refused"),
                httpx.Response(200),
            ]
        )
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2
        assert mock_sleep.call_count == 1

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_exhausted_retries_on_server_error_returns_response(self, mock_sleep):
        mock = MockTransport([httpx.Response(502)] * 4)
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 502
        assert mock.call_count == 4
        assert mock_sleep.call_count == 3

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_exhausted_retries_on_connect_error_raises(self, mock_sleep):
        mock = MockTransport([httpx.ConnectError(f"fail {i}") for i in range(4)])
        transport = RetryTransport(max_retries=3, transport=mock)
        with pytest.raises(SandboxConnectionError, match="4 attempts"):
            transport.handle_request(_make_request())
        assert mock.call_count == 4

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_exhausted_retries_on_429_returns_response(self, mock_sleep):
        mock = MockTransport([httpx.Response(429, headers={"retry-after": "1"})] * 4)
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 429
        assert mock.call_count == 4

    @pytest.mark.parametrize("status", [200, 400, 401, 404, 409, 422, 500])
    def test_no_retry_on_non_retryable_status(self, status):
        mock = MockTransport([httpx.Response(status)])
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == status
        assert mock.call_count == 1

    def test_max_retries_zero_no_retry_on_server_error(self):
        mock = MockTransport([httpx.Response(502)])
        transport = RetryTransport(max_retries=0, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 502
        assert mock.call_count == 1

    def test_max_retries_zero_connect_error_raises_immediately(self):
        mock = MockTransport([httpx.ConnectError("refused")])
        transport = RetryTransport(max_retries=0, transport=mock)
        with pytest.raises(SandboxConnectionError, match="1 attempts"):
            transport.handle_request(_make_request())
        assert mock.call_count == 1

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_backoff_increases_with_attempts(self, mock_sleep):
        mock = MockTransport([httpx.Response(502)] * 3 + [httpx.Response(200)])
        transport = RetryTransport(max_retries=3, transport=mock)
        transport.handle_request(_make_request())
        sleep_times = [call[0][0] for call in mock_sleep.call_args_list]
        assert len(sleep_times) == 3
        for i in range(1, len(sleep_times)):
            assert sleep_times[i] >= sleep_times[i - 1]

    @patch("langsmith.sandbox._transport.time.sleep")
    def test_mixed_errors_then_success(self, mock_sleep):
        mock = MockTransport(
            [
                httpx.ConnectError("refused"),
                httpx.Response(502),
                httpx.Response(200),
            ]
        )
        transport = RetryTransport(max_retries=3, transport=mock)
        response = transport.handle_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 3

    def test_connect_error_preserves_cause(self):
        original = httpx.ConnectError("the cause")
        mock = MockTransport([original])
        transport = RetryTransport(max_retries=0, transport=mock)
        with pytest.raises(SandboxConnectionError) as exc_info:
            transport.handle_request(_make_request())
        assert exc_info.value.__cause__ is original


# ============================================================================
# Async AsyncRetryTransport tests
# ============================================================================


class TestAsyncRetryTransport:
    """Tests for the async AsyncRetryTransport."""

    @pytest.fixture
    def _patch_sleep(self):
        with patch("langsmith.sandbox._transport.asyncio.sleep") as mock:
            yield mock

    @pytest.mark.asyncio
    @pytest.mark.parametrize("status", [502, 503, 504])
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_retries_on_server_error_then_succeeds(self, status):
        mock = MockAsyncTransport([httpx.Response(status), httpx.Response(200)])
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_retries_on_429_with_retry_after(self, _patch_sleep):
        mock = MockAsyncTransport(
            [
                httpx.Response(429, headers={"retry-after": "2"}),
                httpx.Response(200),
            ]
        )
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2
        assert _patch_sleep.call_args[0][0] >= 2.0

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_retries_on_connect_error_then_succeeds(self):
        mock = MockAsyncTransport(
            [
                httpx.ConnectError("Connection refused"),
                httpx.Response(200),
            ]
        )
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 2

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_exhausted_retries_on_server_error_returns_response(self):
        mock = MockAsyncTransport([httpx.Response(502)] * 4)
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 502
        assert mock.call_count == 4

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_exhausted_retries_on_connect_error_raises(self):
        mock = MockAsyncTransport([httpx.ConnectError(f"fail {i}") for i in range(4)])
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        with pytest.raises(SandboxConnectionError, match="4 attempts"):
            await transport.handle_async_request(_make_request())
        assert mock.call_count == 4

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_exhausted_retries_on_429_returns_response(self):
        mock = MockAsyncTransport(
            [httpx.Response(429, headers={"retry-after": "1"})] * 4
        )
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 429
        assert mock.call_count == 4

    @pytest.mark.asyncio
    @pytest.mark.parametrize("status", [200, 400, 404, 500])
    async def test_no_retry_on_non_retryable_status(self, status):
        mock = MockAsyncTransport([httpx.Response(status)])
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == status
        assert mock.call_count == 1

    @pytest.mark.asyncio
    async def test_max_retries_zero_no_retry_on_server_error(self):
        mock = MockAsyncTransport([httpx.Response(502)])
        transport = AsyncRetryTransport(max_retries=0, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 502
        assert mock.call_count == 1

    @pytest.mark.asyncio
    async def test_max_retries_zero_connect_error_raises_immediately(self):
        mock = MockAsyncTransport([httpx.ConnectError("refused")])
        transport = AsyncRetryTransport(max_retries=0, transport=mock)
        with pytest.raises(SandboxConnectionError, match="1 attempts"):
            await transport.handle_async_request(_make_request())
        assert mock.call_count == 1

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("_patch_sleep")
    async def test_mixed_errors_then_success(self):
        mock = MockAsyncTransport(
            [
                httpx.ConnectError("refused"),
                httpx.Response(502),
                httpx.Response(200),
            ]
        )
        transport = AsyncRetryTransport(max_retries=3, transport=mock)
        response = await transport.handle_async_request(_make_request())
        assert response.status_code == 200
        assert mock.call_count == 3

    @pytest.mark.asyncio
    async def test_connect_error_preserves_cause(self):
        original = httpx.ConnectError("the cause")
        mock = MockAsyncTransport([original])
        transport = AsyncRetryTransport(max_retries=0, transport=mock)
        with pytest.raises(SandboxConnectionError) as exc_info:
            await transport.handle_async_request(_make_request())
        assert exc_info.value.__cause__ is original



import os
import statistics
import time
from pathlib import Path
from typing import Dict

from langsmith import Client
from langsmith.schemas import ExampleUpsertWithAttachments

WRITE_BATCH = 10000


def create_large_file(size: int, dir: str) -> str:
    """Create a large file for benchmarking purposes."""
    filename = f"large_file_{size}.txt"
    filepath = os.path.join(dir, filename)

    # delete the file if it exists
    print("Deleting existing file...")
    if os.path.exists(filepath):
        os.remove(filepath)

    print("Creating big file...")
    with open(filepath, "w") as f:
        curr_size = 0
        while curr_size < size:
            f.write("a" * (size - curr_size))
            curr_size += size - curr_size

    print("Done creating big file...")
    return filepath


DATASET_NAME = "upsert_big_file_to_dataset"


def benchmark_big_file_upload(
    size_bytes: int, num_examples: int, samples: int = 1
) -> Dict:
    """
    Benchmark run creation with specified parameters.
    Returns timing statistics.
    """
    multipart_timings = []

    for _ in range(samples):
        client = Client()

        if client.has_dataset(dataset_name=DATASET_NAME):
            client.delete_dataset(dataset_name=DATASET_NAME)

        dataset = client.create_dataset(
            DATASET_NAME,
            description="Test dataset for big file upload",
        )
        large_file = create_large_file(size_bytes, "/tmp")
        examples = [
            ExampleUpsertWithAttachments(
                dataset_id=dataset.id,
                inputs={"a": 1},
                outputs={"b": 2},
                attachments={
                    "bigfile": ("text/plain", Path(large_file)),
                },
            )
            for _ in range(num_examples)
        ]

        multipart_start = time.perf_counter()
        client.upsert_examples_multipart(upserts=examples)
        multipart_elapsed = time.perf_counter() - multipart_start

        multipart_timings.append(multipart_elapsed)

    return {
        "mean": statistics.mean(multipart_timings),
        "median": statistics.median(multipart_timings),
        "stdev": (
            statistics.stdev(multipart_timings) if len(multipart_timings) > 1 else 0
        ),
        "min": min(multipart_timings),
        "max": max(multipart_timings),
    }


size_bytes = 50000000
num_examples = 10


def main(size_bytes: int, num_examples: int = 1):
    """
    Run benchmarks with different combinations of parameters and report results.
    """
    results = benchmark_big_file_upload(size_bytes, num_examples)

    print(f"\nBenchmark Results for size {size_bytes} and {num_examples} examples:")
    print("-" * 30)
    print(f"{'Metric':<15} {'Result':>20}")
    print("-" * 30)

    metrics = ["mean", "median", "stdev", "min", "max"]
    for metric in metrics:
        print(f"{results[metric]:>20.4f}")

    print("-" * 30)
    print(f"{'Throughput':<15} {num_examples / results['mean']:>20.2f} ")
    print("(examples/second)")


if __name__ == "__main__":
    main(size_bytes, num_examples)



from typing import Any, Optional


class AnyInt(int):
    def __init__(self, low: Optional[int] = None, high: Optional[int] = None) -> None:
        """Integer with optional interval [low, high)."""
        super().__init__()
        self._low = low
        self._high = high

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, int):
            return False
        if self._low is not None and other < self._low:
            return False
        if self._high is not None and other >= self._high:
            return False
        return True



import datetime
from typing import Any
from unittest.mock import MagicMock

from langsmith._internal._uuid import uuid7, uuid7_deterministic
from langsmith.client import Client
from langsmith.run_helpers import traceable
from langsmith.run_trees import RunTree


def _uuid_v7_ms(uuid_obj) -> int:
    """Extract milliseconds since epoch from a UUIDv7 using string layout.

    UUIDv7 stores Unix time in ms in the first 12 hex chars of the canonical
    string representation (48 msb bits).
    """
    s = str(uuid_obj).replace("-", "")
    return int(s[:12], 16)


def test_traceable_uses_uuidv7_and_start_time_matches_run_id() -> None:
    captured: dict[str, Any] = {}

    # Use a parent run to ensure a child RunTree is created even if tracing is disabled
    parent = RunTree(name="parent", run_type="chain", inputs={})

    @traceable(run_type="chain")
    def fn(x: int, *, run_tree: RunTree) -> int:  # type: ignore[misc]
        # capture run_tree from injected arg
        captured["run_tree"] = run_tree
        return x + 1

    # Execute traced function with explicit parent
    assert fn(1, langsmith_extra={"run_tree": parent}) == 2

    run_tree: RunTree = captured["run_tree"]
    # UUID version must be 7
    assert run_tree.id.version == 7
    # Extract ms from UUID and compare to start_time
    id_ms = _uuid_v7_ms(run_tree.id)
    # Match the same calculation path used by uuid7(id from start_time)
    start_ms = int(run_tree.start_time.timestamp() * 1_000_000_000) // 1_000_000
    assert abs(id_ms - start_ms) <= 1


def test_run_tree_default_uuidv7_and_start_time_match() -> None:
    # Default RunTree should have v7 id and matching start_time
    rt = RunTree(name="test", run_type="chain", inputs={})

    assert rt.id.version == 7
    id_ms = _uuid_v7_ms(rt.id)
    start_ms = int(rt.start_time.timestamp() * 1_000_000_000) // 1_000_000
    assert abs(id_ms - start_ms) <= 1


def test_run_tree_explicit_start_time_matches_id() -> None:
    fixed_start = datetime.datetime(
        2024, 6, 7, 8, 9, 10, 123000, tzinfo=datetime.timezone.utc
    )
    rt = RunTree(name="test", run_type="chain", inputs={}, start_time=fixed_start)

    assert rt.id.version == 7
    id_ms = _uuid_v7_ms(rt.id)
    start_ms = int(rt.start_time.timestamp() * 1_000_000_000) // 1_000_000
    assert id_ms == start_ms


def test_post_and_patch_include_run_type_and_start_time() -> None:
    client = MagicMock(spec=Client)
    fixed_start = datetime.datetime(
        2024, 6, 7, 8, 9, 10, 123000, tzinfo=datetime.timezone.utc
    )
    rt = RunTree(
        name="post-patch-test",
        run_type="llm",
        inputs={"x": 1},
        start_time=fixed_start,
    )
    # Inject our dummy client (use _client for backwards-compat field)
    rt._client = client  # type: ignore[assignment]

    # Post should include run_type and start_time
    rt.post()
    assert client.create_run.called
    _, create_kwargs = client.create_run.call_args
    assert create_kwargs["run_type"] == "llm"
    assert create_kwargs["start_time"] == fixed_start

    # Patch should include run_type and start_time
    rt.patch()
    assert client.update_run.called
    _, update_kwargs = client.update_run.call_args
    assert update_kwargs["run_type"] == "llm"
    assert update_kwargs["start_time"] == fixed_start


def test_uuid7_deterministic_produces_expected_values() -> None:
    """Test with hard-coded values to ensure compatibility across implementations.

    These values should match the TypeScript implementation exactly.
    """
    from uuid import UUID

    test_cases = [
        {
            "input": UUID("019c0711-e1aa-7223-bf21-12119afe80f7"),
            "key": "other-key",
            "expected": UUID("019c0711-e1aa-7614-833c-500ba8e2e686"),
        },
        {
            "input": UUID("019c0711-e1aa-7223-bf21-12119afe80f7"),
            "key": "replica-project",
            "expected": UUID("019c0711-e1aa-7817-8301-943c0df9a3cd"),
        },
        {
            "input": UUID("01900000-0000-7000-8000-000000000000"),
            "key": "test",
            "expected": UUID("01900000-0000-7132-8131-f80e352488a3"),
        },
    ]

    for test_case in test_cases:
        result = uuid7_deterministic(test_case["input"], test_case["key"])
        assert result == test_case["expected"], (
            f"Expected {test_case['expected']} but got {result} "
            f"for input {test_case['input']} with key '{test_case['key']}'"
        )
        assert result.version == 7


def test_uuid7_deterministic() -> None:
    """Test uuid7_deterministic produces valid, deterministic UUID7s."""

    original = uuid7()
    key = "replica-project"

    d1 = uuid7_deterministic(original, key)
    d2 = uuid7_deterministic(original, key)

    # Valid UUID7
    assert d1.version == 7
    assert (d1.bytes[8] & 0xC0) == 0x80  # RFC 4122 variant

    assert d1 == d2

    # Different inputs -> different outputs
    assert uuid7_deterministic(original, "other-key") != d1
    assert uuid7_deterministic(uuid7(), key) != d1


def test_uuid7_deterministic_timestamp_handling() -> None:
    """Test timestamp is preserved from UUID7, fresh for non-UUID7."""
    import time
    import uuid as uuid_module

    # UUID7 input: timestamp preserved
    original_v7 = uuid7()
    derived_v7 = uuid7_deterministic(original_v7, "key")
    assert _uuid_v7_ms(derived_v7) == _uuid_v7_ms(original_v7)

    # UUID4 input: gets fresh timestamp
    before_ms = time.time_ns() // 1_000_000
    derived_v4 = uuid7_deterministic(uuid_module.uuid4(), "key")
    after_ms = time.time_ns() // 1_000_000

    assert derived_v4.version == 7
    assert before_ms <= _uuid_v7_ms(derived_v4) <= after_ms



from unittest import mock

from langsmith import expect
from langsmith._expect import ls_client


def _is_none(x: object) -> bool:
    return x is None


@mock.patch.object(ls_client, "Client", autospec=True)
def test_expect_explicit_none(mock_client: mock.Mock) -> None:
    expect(None).against(_is_none)
    expect(None).to_be_none()
    expect.score(1).to_equal(1)
    expect.score(1).to_be_less_than(2)
    expect.score(1).to_be_greater_than(0)
    expect.score(1).to_be_between(0, 2)
    expect.score(1).to_be_approximately(1, 2)
    expect({1, 2}).to_contain(1)



"""Tests for AsyncSandbox class."""

import pytest
from pytest_httpx import HTTPXMock

from langsmith.sandbox import (
    AsyncSandboxClient,
    DataplaneNotConfiguredError,
    ResourceNotFoundError,
    SandboxConnectionError,
    SandboxNotReadyError,
)
from langsmith.sandbox._async_sandbox import AsyncSandbox


@pytest.fixture
async def client():
    """Create an AsyncSandboxClient with retries disabled for test isolation."""
    async with AsyncSandboxClient(
        api_endpoint="http://test-server:8080", max_retries=0
    ) as c:
        yield c


@pytest.fixture
def sandbox(client: AsyncSandboxClient):
    """Create an AsyncSandbox instance."""
    return AsyncSandbox.from_dict(
        data={
            "name": "test-sandbox",
            "template_name": "test-template",
            "dataplane_url": "https://sandbox-router.example.com/sb-123",
        },
        client=client,
        auto_delete=False,
    )


class TestAsyncSandboxProperties:
    """Tests for AsyncSandbox properties."""

    def test_name_property(self, sandbox):
        """Test name property."""
        assert sandbox.name == "test-sandbox"

    def test_template_name_property(self, sandbox):
        """Test template_name property."""
        assert sandbox.template_name == "test-template"

    def test_dataplane_url_property(self, sandbox):
        """Test dataplane_url property."""
        assert sandbox.dataplane_url == "https://sandbox-router.example.com/sb-123"


class TestAsyncSandboxStatusFields:
    """Tests for status fields on AsyncSandbox."""

    def test_from_dict_with_status(self, client):
        """Test from_dict parses status fields."""
        sb = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "status": "provisioning",
                "status_message": None,
                "dataplane_url": "https://sandbox-router.example.com/sb-123",
            },
            client=client,
            auto_delete=False,
        )
        assert sb.status == "provisioning"
        assert sb.status_message is None

    def test_from_dict_defaults_to_ready(self, client):
        """Test from_dict defaults status to 'ready' when absent."""
        sb = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
            },
            client=client,
            auto_delete=False,
        )
        assert sb.status == "ready"
        assert sb.status_message is None

    async def test_provisioning_sandbox_blocks_run(self, client):
        """Test that run() raises SandboxNotReadyError for non-ready sandbox."""
        sb = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "status": "provisioning",
                "dataplane_url": "https://sandbox-router.example.com/sb-123",
            },
            client=client,
            auto_delete=False,
        )
        with pytest.raises(SandboxNotReadyError, match="not ready"):
            await sb.run("echo hello")

    async def test_provisioning_sandbox_blocks_write(self, client):
        """Test that write() raises SandboxNotReadyError for non-ready sandbox."""
        sb = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "status": "provisioning",
                "dataplane_url": "https://sandbox-router.example.com/sb-123",
            },
            client=client,
            auto_delete=False,
        )
        with pytest.raises(SandboxNotReadyError):
            await sb.write("/tmp/test.txt", "hello")

    async def test_provisioning_sandbox_blocks_read(self, client):
        """Test that read() raises SandboxNotReadyError for non-ready sandbox."""
        sb = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "status": "provisioning",
                "dataplane_url": "https://sandbox-router.example.com/sb-123",
            },
            client=client,
            auto_delete=False,
        )
        with pytest.raises(SandboxNotReadyError):
            await sb.read("/tmp/test.txt")


class TestAsyncSandboxRun:
    """Tests for async sandbox run command."""

    async def test_run_command_success(self, sandbox, httpx_mock: HTTPXMock):
        """Test running a successful command."""
        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={
                "stdout": "hello world\n",
                "stderr": "",
                "exit_code": 0,
            },
        )

        result = await sandbox.run("echo hello world")

        assert result.stdout == "hello world\n"
        assert result.stderr == ""
        assert result.exit_code == 0
        assert result.success is True

    async def test_run_command_failure(self, sandbox, httpx_mock: HTTPXMock):
        """Test running a failing command."""
        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={
                "stdout": "",
                "stderr": "command not found\n",
                "exit_code": 127,
            },
        )

        result = await sandbox.run("nonexistent-command")

        assert result.exit_code == 127
        assert result.success is False

    async def test_run_connection_error(self, sandbox, httpx_mock: HTTPXMock):
        """Test run with connection error."""
        import httpx

        httpx_mock.add_exception(httpx.ConnectError("Connection refused"))

        with pytest.raises(SandboxConnectionError):
            await sandbox.run("echo hello")

    async def test_run_without_dataplane_url(self, client: AsyncSandboxClient):
        """Test run raises error when dataplane_url is not configured."""
        sandbox = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "dataplane_url": None,
            },
            client=client,
            auto_delete=False,
        )

        with pytest.raises(DataplaneNotConfiguredError) as exc_info:
            await sandbox.run("echo hello")

        assert "test-sandbox" in str(exc_info.value)
        assert "dataplane_url" in str(exc_info.value)

    async def test_run_with_env(self, sandbox, httpx_mock: HTTPXMock):
        """Test running a command with environment variables."""
        import json

        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={"stdout": "hello\n", "stderr": "", "exit_code": 0},
        )

        await sandbox.run("echo $MY_VAR", env={"MY_VAR": "hello", "OTHER": "value"})

        request = httpx_mock.get_request()
        payload = json.loads(request.content)
        assert payload["env"] == {"MY_VAR": "hello", "OTHER": "value"}

    async def test_run_with_cwd(self, sandbox, httpx_mock: HTTPXMock):
        """Test running a command with custom working directory."""
        import json

        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={"stdout": "/tmp\n", "stderr": "", "exit_code": 0},
        )

        await sandbox.run("pwd", cwd="/tmp")

        request = httpx_mock.get_request()
        payload = json.loads(request.content)
        assert payload["cwd"] == "/tmp"

    async def test_run_with_custom_shell(self, sandbox, httpx_mock: HTTPXMock):
        """Test running a command with custom shell."""
        import json

        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={"stdout": "", "stderr": "", "exit_code": 0},
        )

        await sandbox.run("echo hello", shell="/bin/sh")

        request = httpx_mock.get_request()
        payload = json.loads(request.content)
        assert payload["shell"] == "/bin/sh"

    async def test_run_default_shell(self, sandbox, httpx_mock: HTTPXMock):
        """Test that default shell is /bin/bash."""
        import json

        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={"stdout": "", "stderr": "", "exit_code": 0},
        )

        await sandbox.run("echo hello")

        request = httpx_mock.get_request()
        payload = json.loads(request.content)
        assert payload["shell"] == "/bin/bash"

    async def test_run_omits_none_values(self, sandbox, httpx_mock: HTTPXMock):
        """Test that None values for env and cwd are omitted from payload."""
        import json

        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/execute",
            json={"stdout": "", "stderr": "", "exit_code": 0},
        )

        await sandbox.run("echo hello", env=None, cwd=None)

        request = httpx_mock.get_request()
        payload = json.loads(request.content)
        assert "env" not in payload
        assert "cwd" not in payload


class TestAsyncSandboxWrite:
    """Tests for async sandbox file write."""

    async def test_write_text_file(self, sandbox, httpx_mock: HTTPXMock):
        """Test writing a text file."""
        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/upload?path=%2Fapp%2Ftest.txt",
            json={
                "path": "/app/test.txt",
                "written": 11,
            },
        )

        # Should not raise
        await sandbox.write("/app/test.txt", "hello world")

        # Verify request is multipart form
        request = httpx_mock.get_request()
        assert b"hello world" in request.content
        content_type = request.headers.get("content-type", "")
        assert content_type.startswith("multipart/form-data")

    async def test_write_binary_file(self, sandbox, httpx_mock: HTTPXMock):
        """Test writing a binary file."""
        httpx_mock.add_response(
            method="POST",
            url="https://sandbox-router.example.com/sb-123/upload?path=%2Fapp%2Fdata.bin",
            json={
                "path": "/app/data.bin",
                "written": 4,
            },
        )

        binary_data = b"\x00\x01\x02\x03"
        await sandbox.write("/app/data.bin", binary_data)

        # Verify request is multipart form with binary content
        request = httpx_mock.get_request()
        assert binary_data in request.content
        content_type = request.headers.get("content-type", "")
        assert content_type.startswith("multipart/form-data")

    async def test_write_without_dataplane_url(self, client: AsyncSandboxClient):
        """Test write raises error when dataplane_url is not configured."""
        sandbox = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "dataplane_url": None,
            },
            client=client,
            auto_delete=False,
        )

        with pytest.raises(DataplaneNotConfiguredError) as exc_info:
            await sandbox.write("/app/test.txt", "hello")

        assert "test-sandbox" in str(exc_info.value)


class TestAsyncSandboxRead:
    """Tests for async sandbox file read."""

    async def test_read_text_file(self, sandbox, httpx_mock: HTTPXMock):
        """Test reading a text file."""
        # URL uses query parameter with encoded path
        httpx_mock.add_response(
            method="GET",
            url="https://sandbox-router.example.com/sb-123/download?path=%2Fapp%2Ftest.txt",
            content=b"hello world",
            headers={"Content-Type": "application/octet-stream"},
        )

        content = await sandbox.read("/app/test.txt")

        assert content == b"hello world"

    async def test_read_binary_file(self, sandbox, httpx_mock: HTTPXMock):
        """Test reading a binary file."""
        binary_data = b"\x00\x01\x02\x03"

        httpx_mock.add_response(
            method="GET",
            url="https://sandbox-router.example.com/sb-123/download?path=%2Fapp%2Fdata.bin",
            content=binary_data,
            headers={"Content-Type": "application/octet-stream"},
        )

        content = await sandbox.read("/app/data.bin")

        assert content == binary_data

    async def test_read_file_not_found(self, sandbox, httpx_mock: HTTPXMock):
        """Test reading non-existent file."""
        httpx_mock.add_response(
            method="GET",
            url="https://sandbox-router.example.com/sb-123/download?path=%2Fapp%2Fnonexistent.txt",
            json={"error": "NotFound", "message": "file not found"},
            status_code=404,
        )

        with pytest.raises(ResourceNotFoundError):
            await sandbox.read("/app/nonexistent.txt")

    async def test_read_without_dataplane_url(self, client: AsyncSandboxClient):
        """Test read raises error when dataplane_url is not configured."""
        sandbox = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
                "dataplane_url": None,
            },
            client=client,
            auto_delete=False,
        )

        with pytest.raises(DataplaneNotConfiguredError) as exc_info:
            await sandbox.read("/app/test.txt")

        assert "test-sandbox" in str(exc_info.value)


class TestAsyncSandboxContextManager:
    """Tests for async sandbox context manager."""

    async def test_auto_delete_on_exit(
        self, client: AsyncSandboxClient, httpx_mock: HTTPXMock
    ):
        """Test sandbox is deleted on async context exit."""
        httpx_mock.add_response(
            method="DELETE",
            url="http://test-server:8080/boxes/test-sandbox",
            status_code=204,
        )

        sandbox = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
            },
            client=client,
            auto_delete=True,
        )

        async with sandbox:
            pass

        # Verify delete was called
        requests = httpx_mock.get_requests()
        assert len(requests) == 1
        assert requests[0].method == "DELETE"

    async def test_no_delete_when_auto_delete_false(
        self, client: AsyncSandboxClient, httpx_mock: HTTPXMock
    ):
        """Test sandbox is not deleted when auto_delete is False."""
        sandbox = AsyncSandbox.from_dict(
            data={
                "name": "test-sandbox",
                "template_name": "test-template",
            },
            client=client,
            auto_delete=False,
        )

        async with sandbox:
            pass

        # Verify no delete request
        requests = httpx_mock.get_requests()
        assert len(requests) == 0
